      SUBROUTINE DEXPTAYNS3( A, N, KMAX, WORK, LWORK, INFO )
      IMPLICIT NONE

*     .. Scalar Arguments ..
      INTEGER            N, KMAX, LWORK, INFO 
*     .. Array Arguments ..
      DOUBLE PRECISION   A(N,*), WORK(*) 
*    ..
*
*  Purpose
*  =======
*
*  To compute matrix exponential B = exp(A) by means of Taylor series 
*  with scaling and squaring.
*
*  Parameters
*  ==========
*
*  A      - DOUBLE PRECISION array, dimension (N,N)
*           On entry, A contains the matrix to be computed its exponential.
*           On exit, A contains the matrix exponential.
*
*  N      - INTEGER 
*           On entry, N represents the size of A.
*           Unchanged on exit 
*
*  KMAX   - INTEGER 
*           On entry, KMAX represents the maximum order to be used for the
*           approximation:
*
*                                       KMAX:  6  7  8  9
*             Maximum approximation order, M: 16 20 25 30
*
*           If KMAX<6 or KMAX>9, KMAX will be set to 9.
*           Note that the best order approximation depends on the matrix
*           norm and it is always determined automatically. KMAX only
*           stablishes the maximum of this order.
*
*  WORK   - (workspace) DOUBLE PRECISION array, dimension (LWORK)
*           On exit, if INFO = 0, 
*              If LWORK = -1 on entry
*                   WORK(1) returns the optimal LWORK.
*              else
*                   WORK(1) returns the scaling applied (S). 
*                   WORK(2) returns the approximation order used (M).
*                   WORK(3) returns the total matrix products (CP).
*
*  LWORK  - INTEGER
*           On entry, the dimension of the array WORK. 
*           LWORK >= max(1, (Q+1)*N*N), where Q = FLOOR( SQRT(M) )
*
*           If LWORK = -1, then a workspace query is assumed; the
*           routine only calculates the optimal size of the WORK array,
*           returns this value as the first entry of the WORK array, 
*           and no error message related to LWORK is issued.
*
*  INFO   - INTEGER
*           = 0:  successful exit
*           < 0:  if INFO = -i, the i-th argument had an illegal value
*
*  Further Details
*  ===============
*
*   References:
*
*   [1] J. Sastre, J. Ibez, E. Defez, P. Ruiz, Computing matrix 
*       exponential to solve coupled differential models in Engineering,
*       in Proc. of the International Conference on Mathematical Models
*       of Addictive Behaviour, Medicine & Engineering 2010 (I.S.B.N.:
*       978-84-693-9537-0), Valencia, September 8th-10th, p. 193-196. 
*   
*   [2] J. Sastre, J. Ibez, E. Defez, P. Ruiz, Accurate matrix 
*       exponential computation to solve coupled differential models 
*       in Engineering, Mathematical and Computer Modelling, 54(2011),
*       1835-1840, doi:10.1016/j.mcm.2010.12.049.
*
*   Last update: 01/07/2014 
*
*  =====================================================================
*
*  .. Local Parameters ..
      DOUBLE PRECISION ZERO, ONE, TWO
      PARAMETER        ( ZERO=0.0D0, ONE=1.0D0, TWO=2.0D0 )
*
*  .. Local Scalars ..
      LOGICAL          LQUERY, CONT
      INTEGER          N2, M, Q, S, LWKMIN, I, J, MW, CP, INM, IAUX,
     $                 MMAX, QMAX
      DOUBLE PRECISION THETAM1, NORM, L, B, DUMMY, DAUX
*
*  .. Local Arrays..
      INTEGER          WI(6), IP(5,2), MV(9), LT(8), QV(9)
      DOUBLE PRECISION THETA(9), UCM2(9), C(9), NA(32)
*
*  .. Intrinsic Functions ..
      INTRINSIC        SQRT, LOG, FLOOR, CEILING, MIN, MAX, ABS
*
*  .. External Subroutines ..
      EXTERNAL         DCOPY, DGEMM, DAPBI, MTPS
*     
*  .. External Functions ..
      DOUBLE PRECISION DLANGE, NORM1PP, NORM1P
      EXTERNAL         DLANGE, NORM1PP, NORM1P
*
*     ..
*     .. Executable Statements ..

      DATA IP / 1, 1, 2, 1, 2,
     $          1, 2, 2, 4, 4 /

      DATA MV / 2, 4, 6, 9, 12, 16, 20, 25, 30 /

      DATA LT / 8, 11, 14, 18, 22, 27, 32, 37 /

      DATA QV / 2, 2, 3, 3, 4, 4, 5, 5, 5 /

      THETAM1  = 1.490116111983279D-8
      THETA(1) = 8.733457513635361D-6 
      THETA(2) = 1.678018844321752D-3
      THETA(3) = 1.773082199654024D-2
      THETA(4) = 1.137689245787824D-1
      THETA(5) = 3.280542018037257D-1
      THETA(6) = 7.912740176600240D-1
      THETA(7) = 1.438252596804337
      THETA(8) = 2.428582524442827
      THETA(9) = 3.539666348743690

      UCM2(1) = 8.9D-16
      UCM2(2) = 1.598721155460225D-14
      UCM2(3) = 6.394884621840902D-13
      UCM2(4) = 4.431655042935745D-10
      UCM2(5) = 7.445180472132051D-07
      UCM2(6) = 4.181213353149360D-02
      UCM2(7) = 5.942340417495871D+03
      UCM2(8) = 4.649643683073819D+10
      UCM2(9) = 9.423674633957229D+17

      C(1) = 4D0/3D0
      C(2) = 6D0/5D0
      C(3) = 8D0/7D0
      C(4) = 11D0/10D0
      C(5) = 14D0/13D0
      C(6) = 18D0/17D0
      C(7) = 22D0/21D0
      C(8) = 27D0/26D0
      C(9) = 32D0/31D0

      IF ( KMAX.LT.6 .OR. KMAX.GT.9 ) KMAX=9

      IF ( KMAX.LT.8 ) QV(7) = 4
*
*     Test the input arguments
*
      N2 = N*N
      INFO = 0
      LQUERY = ( LWORK.EQ.-1 )

      Q = QV(KMAX)
      LWKMIN = N2*Q+N2
      MW = N2*Q+1

      IF ( N.LT.0 ) THEN
         INFO = -2
      ELSE IF ( LWORK.LT.LWKMIN .AND. .NOT.LQUERY ) THEN
         INFO =-5
      END IF      

      IF ( INFO.LT.0 ) THEN
         WRITE( *, FMT = 9999 ) 'DEXPTAYNS3', INFO
         STOP
      ELSE IF ( LQUERY ) THEN
         WORK(1) = LWKMIN
         RETURN
      END IF
*
*     Quick return if possible
*
      CP = 0
      S = 0
      NA(1) = DLANGE( '1', N, N, A, N, DUMMY )
      IF ( NA(1).LE.THETAM1 ) THEN
C        F = A + I
         CALL DAPBI( A, N, ONE )
         M = 1
         GOTO 90
      END IF

C     Copy A in WORK(1)
      CALL DCOPY(N2, A, 1, WORK(1), 1 )
      WI(1) = 1
C     Flags for later
      NA(27) = -ONE
      NA(32) = -ONE

      DO INM = 1, KMAX
         M = MV(INM)
         Q = QV(INM) 

         IF ( M.EQ.2 ) THEN

C           Compute A^2 and save in WORK
            WI(2) = WI(1)+N2
            CALL DGEMM( 'N', 'N', N, N, N, ONE, WORK( WI(1) ), N,
     $                  WORK( WI(1) ), N, ZERO, WORK( WI(2) ), N )
            CP = CP+1

            NA(3) = NORM1PP( N, WORK( WI(2) ), 1, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(1)
            DAUX =  C(1)*NA(3) 
            IF ( DAUX.LE.B ) THEN
               NA(4) = NORM1P( N, WORK( WI(2) ), 2 )
               IF ( DAUX+NA(4).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF 
            END IF

         ELSE IF ( M.EQ.4 ) THEN

            NA(5) = NORM1PP( N, WORK( WI(2) ), 2, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(2)
            DAUX =  C(2)*NA(5)
            IF ( DAUX.LE.B ) THEN
               NA(6) = NORM1P( N, WORK( WI(2) ), 3 )
               IF ( DAUX+NA(6).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF

         ELSE IF ( M.EQ.6 ) THEN

C           Compute A^3 and save in WORK
            WI(3) = WI(2)+N2
            CALL DGEMM( 'N', 'N', N, N, N, ONE, WORK( WI(2) ), N,
     $                  WORK( WI(1) ), N, ZERO, WORK( WI(3) ), N )
            CP = CP+1

            NA(7) = NORM1PP( N, WORK( WI(3) ), 2, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(3)
            DAUX =  C(3)*NA(7)
            IF ( DAUX.LE.B ) THEN
               NA(8) = NORM1PP( N, WORK( WI(3) ), 2, WORK( WI(2) ), 
     $                          WORK(MW) )
               IF ( DAUX+NA(8).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF

         ELSE IF ( M.EQ.9 ) THEN

            NA(10) = NORM1PP( N, WORK( WI(3) ), 3, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(4)
            DAUX =  C(4)*NA(10)
            IF ( DAUX.LE.B ) THEN
               NA(11) = NORM1PP( N, WORK( WI(3) ), 3, WORK( WI(2) ),
     $                           WORK(MW) )
               IF ( DAUX+NA(11).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF

         ELSE IF ( M.EQ.12 ) THEN

C           Compute A^4 and save in WORK
            WI(4) = WI(3)+N2
            CALL DGEMM( 'N', 'N', N, N, N, ONE, WORK( WI(2) ), N,
     $                  WORK( WI(2) ), N, ZERO, WORK( WI(4) ), N )
            CP = CP+1

            NA(13) = NORM1PP( N, WORK( WI(4) ), 3, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(5)
            DAUX =  C(5)*NA(13)
            IF ( DAUX.LE.B ) THEN
               NA(14) = NORM1PP( N, WORK( WI(4) ), 3, WORK( WI(2) ),
     $                          WORK(MW) )
               IF ( DAUX+NA(14).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF

         ELSE IF ( M.EQ.16 ) THEN

            NA(17) = NORM1PP( N, WORK( WI(4) ), 4, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(6)
            DAUX =  C(6)*NA(17)
            IF ( DAUX.LE.B ) THEN
               NA(18) = NORM1PP( N, WORK( WI(4) ), 4, WORK( WI(2) ),
     $                           WORK(MW) )
               IF ( DAUX+NA(18).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF

         ELSE IF ( M.EQ.20 ) THEN

C           Compute A^5 and save in WORK
            WI(5) = WI(4)+N2
            CALL DGEMM( 'N', 'N', N, N, N, ONE, WORK( WI(4) ), N,
     $                  WORK( WI(1) ), N, ZERO, WORK( WI(5) ), N )
            CP = CP+1

            NA(21) = NORM1PP( N, WORK( WI(5) ), 4, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(7)
            DAUX =  C(7)*NA(21)
            IF ( DAUX.LE.B ) THEN
               NA(22) = NORM1PP( N, WORK( WI(5) ), 4, WORK( WI(2) ),
     $                          WORK(MW) )
               IF ( DAUX+NA(22).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF

         ELSE IF ( M.EQ.25 ) THEN

            NA(26) = NORM1PP( N, WORK( WI(5) ), 5, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(8)
            DAUX =  C(8)*NA(26)
            IF ( DAUX.LE.B ) THEN
               NA(27) = NORM1PP( N, WORK( WI(5) ), 5, WORK( WI(2) ),
     $                           WORK(MW) )
               IF ( DAUX+NA(27).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF

         ELSE IF ( M.EQ.30 ) THEN

            NA(31) = NORM1PP( N, WORK( WI(5) ), 6, A, WORK(MW) )
            B = MAX( ONE, NA(1) )*UCM2(9)
            DAUX =  C(9)*NA(31)
            IF ( DAUX.LE.B ) THEN
               NA(32) = NORM1PP( N, WORK( WI(5) ), 6, WORK( WI(2) ),
     $                           WORK(MW) )
               IF ( DAUX+NA(32).LE.B ) THEN
                  CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
                  GOTO 90
               END IF
            END IF
         END IF

      END DO

      MMAX = MV(KMAX)
      QMAX = QV(KMAX)
      IAUX = MMAX+2

C     Estimate 1-norm of A^(MMAX+2) if it is not already estimated
      IF ( NA(IAUX).LT.ZERO ) THEN
         NA(IAUX) = NORM1PP( N, WORK( WI(QMAX) ), IAUX/QMAX, 
     $                    WORK( WI(MOD(IAUX,QMAX)) ), WORK(MW) )
      END IF

C     Compute alpha_min
      NORM = MAX( NA(MMAX+1)**(ONE/(MMAX+1)), NA(IAUX)**(ONE/IAUX) )

      L = LOG( NORM/THETA(KMAX) )/LOG( TWO )
      S = CEILING( L )
C     Adjust S if norm/Theta(kmax) is a power of 2
      IF( S.EQ.L ) THEN
         S = S-1
      END IF

C     Test if s can be reduced, checking if (5) of [2] holds with s=s-1
      IF ( S.GE.1 )
         S = S-1
         B = MAX( ONE, NA(1)/(2**S) )*UCM2(KMAX)
         DAUX = C(KMAX)*NA(MMAX+1)/(2**((MMAX+1)*S)) + 
     $          NA(IAUX)/(2**(IAUX*S))
         IF ( DAUX.GT.B ) THEN
            S = S+1
         END IF
      END IF

C     Test if scaled matrix allows using MV(KMAX-1)
      MMAX = MV(KMAX-1)
      QMAX = QV(KMAX-1)
      IAUX = MMAX+2

C     Estimate 1-norm of A^(MMAX+2) if it is not already estimated
      IF ( NA(IAUX).LT.ZERO ) THEN
         NA(IAUX) = NORM1PP( N, WORK( WI(QMAX) ), IAUX/QMAX, 
     $                    WORK( WI(MOD(IAUX,QMAX)) ), WORK(MW) )
      END IF

      B = MAX( ONE, NA(1)/(2**S) )*UCM2(KMAX-1)
      DAUX = C(KMAX-1)*NA(MMAX+1)/(2**((MMAX+1)*S)) + 
     $       NA(IAUX)/(2**(IAUX*S))
      IF ( DAUX.LE.B ) THEN
C        Scaled matrix allows using MV(KMAX-1)      
         M = MV(KMAX-1)
      END IF
           
      CALL MTPS( A, N, WORK, WI, M, Q, S, CP, WORK(MW) )
      
  90  WORK(1) = S
      WORK(2) = M
      WORK(3) = CP

      RETURN

 9999 FORMAT( ' ** On entry to ', A10, ' parameter number ', I2, 
     $        ' had ', 'an illegal value' )
*
*     End of DEXPTAYNS3
*
      END
