      SUBROUTINE MTPS( F, N, AP, IAP, M, Q, S, CP, WORK )
      IMPLICIT NONE

*     .. Scalar Arguments ..
      INTEGER            N, M, Q, S, CP
*     .. Array Arguments ..
      INTEGER            IAP( * )
      DOUBLE PRECISION   F( N,* ), AP( * ), WORK( * )
*    ..
*
*  Purpose
*  =======
*
*  To compute the exponential of matrix F by means of scaling and squaring
*  Taylor algorithm.
*
*  Parameters
*  ==========
*
*  F      - DOUBLE PRECISION array, dimension (N,N)
*           On entry, F contains the matrix to be computed its exponential.
*           On exit, F contains the matrix exponential.
*
*  N      - INTEGER 
*           On entry, N represents the size of F.
*           Unchanged on exit 
*
*  AP     - DOUBLE PRECISION array, dimension Q*N*N
*           On entry, AP contains the matrix powers:
*
*                               F^i, i = 1, 2, ... Q
*           
*  IAP    - INTEGER array, dimension Q
*           On entry, IAP(i) contains the index of array AP, where 
*           F^i begins. 
*           Unchanged on exit 
*
*  M      - INTEGER
*           Taylor approximation order. 
*           Unchanged on exit. 
*
*  Q      - INTEGER
*           Unchanged on exit. 
*        
*  S      - INTEGER
*           Scaling parameter.
*           Unchanged on exit. 
*        
*  CP     - INTEGER
*           Matrix products counter.
*           On exit, CP is updated. 
*        
*  WORK   - (workspace) DOUBLE PRECISION array, dimension (N*N), at
*           least.
*
*  Further Details
*  ===============
*
*  A modified Taylor Paterson-Stockmeyer algorithm for computing the
*  exponential of a matrix is used (see (10) of [1]).
*
%  [1] J. Sastre, J. Ibez, E. Defez, P. Ruiz, Accurate matrix 
%      exponential computation to solve coupled differential models 
%      in Engineering, Mathematical and Computer Modelling, 54(2011),
%      1835-1840, doi:10.1016/j.mcm.2010.12.049.
*
*  =====================================================================
*
*     .. Local Parameters ..
      DOUBLE PRECISION ZERO, ONE, TWO
      PARAMETER        ( ZERO = 0.0D+0, ONE = 1.0D+0, TWO = 2.0D+0 )
*
*     .. Local Scalars ..
      INTEGER          I, J, N2
      DOUBLE PRECISION S2, DAUX, TEMP
*
*     .. Local Arrays..
      DOUBLE PRECISION P(42)
*     
*     .. External Subroutines ..
      EXTERNAL         DGEMM, DCOPY, DSCAL, DAXPY, SETFACT, DAPBI
*     
*     ..
*     .. Executable Statements ..
*     
      N2 = N*N

C     Set factorials
      CALL SETFACT( P, 42, M)
      
      IF ( S.GT.0 ) THEN
         S2 = TWO**S
         TEMP = ONE/(S2**Q)
         DO I = M, Q, -Q 
            IF ( I.LT.M ) THEN
               CALL DGEMM( 'N', 'N', N, N, N, TEMP, F, N,
     $                     AP( IAP(Q) ), N, ZERO, WORK, N )
               CP = CP+1
               CALL DCOPY( N2, WORK, 1, F, 1 )
            ELSE
               CALL DCOPY( N2, AP( IAP(Q) ), 1, F, 1 )
               DAUX = TEMP/P(I)
               CALL DSCAL( N2, DAUX, F, 1 )
            END IF
         
            DO J = Q-1, 1, -1
               DAUX = ONE/(P(I-Q+J)*S2**J)
               CALL DAXPY( N2, DAUX, AP( IAP(J) ), 1, F, 1 )
            END DO
         
            IF( I.GT.Q ) THEN
               DAUX = ONE/P(I-Q)
            ELSE
               DAUX = ONE
            END IF
         
            CALL DAPBI( F, N, DAUX ) 
         END DO

         DO I = 1, S
            CALL DCOPY( N2, F, 1, WORK, 1 )
            CALL DGEMM( 'N', 'N', N, N, N, ONE, WORK, N, WORK,  N, ZERO,
     $                  F, N )
         END DO
         CP = CP+S
      ELSE
         DO I = M, Q, -Q 
            IF ( I.LT.M ) THEN
               CALL DGEMM( 'N', 'N', N, N, N, ONE, F, N,
     $                     AP( IAP(Q) ), N, ZERO, WORK, N )
               CP = CP+1
               CALL DCOPY( N2, WORK, 1, F, 1 )
            ELSE
               CALL DCOPY( N2, AP( IAP(Q) ), 1, F, 1 )
               DAUX = ONE/P(I)
               CALL DSCAL( N2, DAUX, F, 1 )
            END IF
         
            DO J = Q-1, 1, -1
               DAUX = ONE/P(I-Q+J)
               CALL DAXPY( N2, DAUX, AP( IAP(J) ), 1, F, 1 )
            END DO
         
            IF( I.GT.Q ) THEN
               DAUX = ONE/P(I-Q)
            ELSE
               DAUX = ONE
            END IF
         
            CALL DAPBI( F, N, DAUX ) 
         END DO
      END IF

*
*     End of MTPS
*
      END



      SUBROUTINE DAPBI( A, N, BETA )
      IMPLICIT NONE

*     .. Scalar Arguments ..
      INTEGER            N
*     .. Array Arguments ..
      DOUBLE PRECISION   A(N,*), BETA
*    ..
*
*  Purpose
*  =======
*
*  To compute A = A + BETA*I
*
*  Parameters
*  ==========
*
*  A      - DOUBLE PRECISION array, dimension (N,N)
*           On entry, A contains the matrix to be added the identity.
*           On exit, A contains result matrix.
*     
*  N      - INTEGER 
*           On entry, N represents the size of A.
*           Unchanged on exit 
*
*  BETA   - DOUBLE PRECISION
*           On entry, BETA specifies the scalar beta.
*           Unchanged on exit 
*
*  =====================================================================
*
*  .. Local Parameters ..
      DOUBLE PRECISION ZERO
      PARAMETER        ( ZERO = 0.0D+0 )
*
*  .. Local Scalars ..
      INTEGER          I
*
*     ..
*     .. Executable Statements ..
*
*     .. Quick return if possible ..
      IF ( BETA.EQ.ZERO ) RETURN

      DO I= 1, N
         A(I,I) = A(I,I) + BETA
      END DO

*
*     End of DAPBI
*
      END



      SUBROUTINE SETFACT( P, N, M )
      IMPLICIT NONE
      
*     .. Scalar Arguments ..
      INTEGER            N, M
*     .. Array Arguments .. 
      DOUBLE PRECISION   P( N )
*    ..
*
*  Purpose
*  =======
*
*  To set the factorials from 0 to 42 for Taylor serial polynomials.
*
*  Parameters
*  ==========
*
*  P      - DOUBLE PRECISION array, dimension (N)
*           On exit, P contains the coefficients of Hermite serial
*           polynomials.
*     
*  N      - INTEGER 
*           On entry, N represents the size of P.
*           Unchanged on exit 
*
*  M      - INTEGER
*           On entry, M specifies the approximation order.
*           Unchanged on exit 
*
*  =====================================================================
*
*     .. Local Scalars ..
      INTEGER          INFO
*
*     ..
*     .. Executable Statements ..
*
*     .. Check parameters ..
      INFO = 0
      IF( N.LT.M )  INFO = -2
      IF( M.GT.42 ) INFO = -3
      IF( INFO.LT.0 ) THEN
         WRITE( *, FMT = 9999 ) 'SETFACT', INFO
         STOP
      END IF

      P(1) = 1D0
      P(2) = 2D0
      P(3) = 6D0
      P(4) = 24D0
      P(5) = 120D0
      P(6) = 720D0
      P(7) = 5040D0
      P(8) = 40320D0
      P(9) = 362880D0
      P(10) = 3628800D0
      P(11) = 39916800D0
      P(12) = 479001600D0
      P(13) = 6227020800D0
      P(14) = 87178291200D0
      P(15) = 1307674368000D0
      P(16) = 20922789888000D0
      P(17) = 355687428096000D0
      P(18) = 6402373705728000D0
      P(19) = 121645100408832000D0
      P(20) = 2432902008176640000D0
      P(21) = 51090942171709440000D0
      P(22) = 1124000727777607680000D0
      P(22) = 1124000727777607680000D0
      P(23) = 25852016738884976640000D0
      P(24) = 620448401733239439360000D0
      P(25) = 15511210043330985984000000D0
      P(26) = 403291461126605635584000000D0
      P(27) = 10888869450418352160768000000D0
      P(28) = 304888344611713860501504000000D0
      P(29) = 8841761993739701954543616000000D0
      P(30) = 265252859812191058636308480000000D0
      P(31) = 8222838654177922817725562880000000D0
      P(32) = 263130836933693530167218012160000000D0
      P(33) = 8683317618811886495518194401280000000D0
      P(34) = 295232799039604140847618609643520000000D0
      P(35) = 10333147966386144929666651337523200000000D0
      P(36) = 371993326789901217467999448150835200000000D0
      P(37) = 13763753091226345046315979581580902400000000D0
      P(38) = 523022617466601111760007224100074291200000000D0
      P(39) = 20397882081197443358640281739902897356800000000D0
      P(40) = 815915283247897734345611269596115894272000000000D0
      P(41) = 33452526613163807108170062053440751665152000000000D0
      P(42) = 1405006117752879898543142606244511569936384000000000D0

 9999 FORMAT( ' ** On entry to ', A7, ' parameter number ', I2,
     $        ' had an illegal value' )
*
*     End of SETFACT
*
      END




     
