*
*  Functions to estimate 1-norm of a matrix
*
      DOUBLE PRECISION FUNCTION NORM1PP( N, A, P, B, WORK )
      IMPLICIT NONE 
         
*     .. Scalar Arguments ..
      INTEGER            N, P
*     .. Array Arguments .. 
      DOUBLE PRECISION   A(N,*), B(N,*), WORK(*)
*    ..
*     
*  Purpose 
*  ======= 
*
*  NORM1PP estimates the 1-norm of A^P*B, where A and B are square 
*  matrix of the same dimension.
*     
*  Parameters
*  ==========
*        
*  N      - INTEGER 
*           On entry, N represents the size of A and B
*           Unchanged on exit 
*
*  A      - DOUBLE PRECISION array, dimenssion NxN
*           A matrix. 
*           Unchanged on exit 
*           
*  P      - INTEGER
*           Power parameter.
*           Unchanged on exit. 
*
*  B      - DOUBLE PRECISION array, dimenssion NxN
*           B matrix. 
*           Unchanged on exit 
*     
*  WORK   - (workspace) DOUBLE PRECISION array, dimension 3*N
*
*  Further Details
*  ===============
*
*  Based on algorithm 4.1 from:
*
*  [1] FORTRAN Codes for Estimating the One-Norm of a Real or Complex
*      Matrix, with Applications to Condition Estimation.
*      ACM Transactions on Mathematical Software, Vol. 14, No. 4, 
*      pp. 381-396, 1988.
*
*  =====================================================================
*
*  .. Local Parameters ..
      DOUBLE PRECISION ZERO, ONE
      PARAMETER        ( ZERO = 0.0D+0, ONE = 1.0D+0 )
*            
*  .. Local Scalars ..
      INTEGER          I, J, K, IX, IV, IPSI, AUX
      DOUBLE PRECISION DUMMY, DAUX1, DAUX2, OLD_NORM
      LOGICAL          CONT
*
*  .. Intrinsic Functions ..
      INTRINSIC        MOD, DABS
*  
*  .. External Subroutines ..
      EXTERNAL         DGEMV
*     
*  .. External Functions ..
      DOUBLE PRECISION DLANGE
      EXTERNAL         DLANGE
*     
*     ..
*     .. Executable Statements ..
*     
      IX   =      1
      IV   =   IX+N
      IPSI =   IV+N
      
C     Initialize vector X to one/n
      DAUX1 = ONE/N
      DO I = 1, N
         WORK(I) = DAUX1
      END DO

C     v = A^p*B*x
C     Step 1: v = B*x
      CALL DGEMV( 'N', N, N, ONE, B, N, WORK(IX), 1, ZERO, WORK(IV), 1 )

C     Step 2: v = A*v, p times
      DO I = 1, P
C        v=A*v        
         CALL DGEMV ( 'N', N, N, ONE, A, N, WORK(IV), 1, ZERO,
     $                WORK(IX), 1 )
         AUX = IX
         IX = IV
         IV = AUX
      END DO

      NORM1PP = DLANGE( '1', N, 1, WORK(IV), 1, DUMMY )

C     psi = sign(v)      
      DO I = 1, N
         DAUX1 = WORK(IV+I-1)
         IF ( DAUX1.EQ.ZERO ) THEN
            WORK(IPSI+I-1) = ZERO
         ELSEIF ( DAUX1.GT.ZERO ) THEN
            WORK(IPSI+I-1) = ONE
         ELSE
            WORK(IPSI+I-1) = -ONE
         END IF
      END DO

C     x = (A')^p*B'*psi
C     Step 1:  x = B'*psi
      CALL DGEMV( 'T', N, N, ONE, B, N, WORK(IPSI), 1, ZERO,
     $            WORK(IX), 1 )

C     Step 2: x=A'*x. p times
      DO I = 1, P
         CALL DGEMV ( 'T', N, N, ONE, A, N, WORK(IX), 1, ZERO,
     $                WORK(IV), 1 )
         AUX = IX
         IX = IV
         IV = AUX
      END DO

      K = 2
      CONT = .TRUE.
      DO WHILE ( CONT )
C        [*,j] = max(abs(x)) 
         J = 1
         DO I = 2, N
            IF ( DABS( WORK(IX+I-1) ).GT.DABS( WORK(IX+J-1) ) ) J = I
         END DO

C        v = A^p*B*e_j
C        Step 1: v = B*e_j --> v=B(:,J)
         DO I = 1, N
            WORK(IV+I-1) = B(I, J)
         END DO

C        Step 2: v=A*v, p times
         DO I = 1, P
            CALL DGEMV ( 'N', N, N, ONE, A, N, WORK(IV), 1, ZERO,
     $                   WORK(IX), 1 )
            AUX = IX
            IX = IV
            IV = AUX
         END DO

         OLD_NORM = NORM1PP
         NORM1PP = DLANGE( '1', N, 1, WORK(IV), 1, DUMMY )

C        Check if sign(v)=psi
         AUX = 0
         I = 1
         DO WHILE ( I.LE.N .AND. AUX.EQ.0 )
            DAUX1 = WORK(IPSI+I-1)
            DAUX2 = WORK(IV+I-1)
            IF ( (DAUX1.EQ.ZERO .AND. DAUX2.NE.ZERO) .OR.
     $           (DAUX1.EQ.ONE  .AND. DAUX2.LE.ZERO) .OR.
     $           (DAUX1.EQ.-ONE .AND. DAUX2.GE.ZERO) ) THEN
               AUX = 1
            END IF
            I = I+1
         END DO 

         IF ( AUX.EQ.0 .OR. NORM1PP.LE.OLD_NORM ) THEN
            CONT = .FALSE.
         ELSE
C           psi = sign(v)      
            DO I = 1, N
               DAUX1 = WORK(IV+I-1)
               IF ( DAUX1.EQ.ZERO ) THEN
                  WORK(IPSI+I-1) = ZERO
               ELSEIF ( DAUX1.GT.ZERO ) THEN
                  WORK(IPSI+I-1) = ONE
               ELSE
                  WORK(IPSI+I-1) = -ONE
               END IF
            END DO

C           x = (A')^p*B'*psi
C           Step 1:  x = B'*psi
            CALL DGEMV( 'T', N, N, ONE, B, N, WORK(IPSI), 1, ZERO,
     $                  WORK(IX), 1 )

C           Step 2: x = A'*x, p times
            DO I = 1, P
               CALL DGEMV ( 'T', N, N, ONE, A, N, WORK(IX), 1, ZERO,
     $                      WORK(IV), 1 )
               AUX = IX
               IX = IV
               IV = AUX
            END DO

            K = K+1
            DAUX1 = DLANGE( 'M', N, 1, WORK(IX), 1, DUMMY )
            IF ( DAUX1.EQ.WORK(IX+J-1) .OR. K.GT.5 ) CONT = .FALSE.
         END IF
      END DO

      DO I = 1, N
         WORK(IX+I-1) = (ONE + (I-ONE)/(N-ONE)) * (-1)**(I+1)
      END DO

C     x = A^p*B*x
C     Step 1: x = B*x
      CALL DGEMV( 'N', N, N, ONE, B, N, WORK(IX), 1, ZERO,
     $            WORK(IV), 1 )
      AUX = IX
      IX = IV
      IV = AUX

C     Step 2:  x=A*x, p times
      DO I = 1, P
         CALL DGEMV ( 'N', N, N, ONE, A, N, WORK(IX), 1, ZERO,
     $                WORK(IV), 1 )
         AUX = IX
         IX = IV
         IV = AUX
      END DO

      DAUX1 = 2*DLANGE( '1', N, 1, WORK(IX), 1, DUMMY )/(3*N)
      IF ( DAUX1.GT.NORM1PP ) NORM1PP = DAUX1                 
*
*     End of NORM1PP
*
      END



      DOUBLE PRECISION FUNCTION NORM1P( N, A, P )
      IMPLICIT NONE 
         
*     .. Scalar Arguments ..
      INTEGER            N, P
*     .. Array Arguments .. 
      DOUBLE PRECISION   A(N,*)
*    ..
*     
*  Purpose 
*  ======= 
*
*  NORM1P estimates the 1-norm of A^P
*     
*  Parameters
*  ==========
*        
*  N      - INTEGER 
*           On entry, N represents the size of A
*           Unchanged on exit 
*
*  A      - DOUBLE PRECISION array, dimenssion NxN
*           A matrix. 
*           Unchanged on exit 
*           
*  P      - INTEGER
*           Power parameter.
*           Unchanged on exit. 
*
*  Further Details
*  ===============
*
*  Based on norm1pp function and Algorithm 4.1 from:
*
*  [1] FORTRAN Codes for Estimating the One-Norm of a Real or Complex
*      Matrix, with Applications to Condition Estimation.
*      ACM Transactions on Mathematical Software, Vol. 14, No. 4, 
*      pp. 381-396, 1988.
*
*  =====================================================================
*
*  .. Local Parameters ..
      DOUBLE PRECISION ZERO, ONE
      PARAMETER        ( ZERO = 0.0D+0, ONE = 1.0D+0 )
*            
*  .. Local Scalars ..
      INTEGER          I, J, K, IX, IV, IPSI, AUX
      DOUBLE PRECISION DUMMY, DAUX1, DAUX2, OLD_NORM
      LOGICAL          CONT
*
*  .. Local Arrays ..
      DOUBLE PRECISION WORK(3*N)
*
*  .. Intrinsic Functions ..
      INTRINSIC        MOD, DABS
*  
*  .. External Subroutines ..
      EXTERNAL         DGEMV
*     
*  .. External Functions ..
      DOUBLE PRECISION DLANGE
      EXTERNAL         DLANGE
*     
*     ..
*     .. Executable Statements ..
*     
      IX   =      1
      IV   =   IX+N
      IPSI =   IV+N
      
C     Initialize vector X to one/n
      DAUX1 = ONE/N
      DO I = 1, N
         WORK(I) = DAUX1
      END DO

C     v = A^p*x
C     Step 1: v = A*x
      CALL DGEMV( 'N', N, N, ONE, A, N, WORK(IX), 1, ZERO, WORK(IV), 1 )

C     Step 2: v = A*v, p-1 times
      DO I = 1, P-1
C        v=A*v        
         CALL DGEMV ( 'N', N, N, ONE, A, N, WORK(IV), 1, ZERO,
     $                WORK(IX), 1 )
C        v <--> x
         AUX = IX
         IX = IV
         IV = AUX
      END DO

      NORM1P = DLANGE( '1', N, 1, WORK(IV), 1, DUMMY )

C     psi = sign(v)      
      DO I = 1, N
         DAUX1 = WORK(IV+I-1)
         IF ( DAUX1.EQ.ZERO ) THEN
            WORK(IPSI+I-1) = ZERO
         ELSEIF ( DAUX1.GT.ZERO ) THEN
            WORK(IPSI+I-1) = ONE
         ELSE
            WORK(IPSI+I-1) = -ONE
         END IF
      END DO

C     x = (A')^p*psi
C     Step 1:  x = A'*psi
      CALL DGEMV( 'T', N, N, ONE, A, N, WORK(IPSI), 1, ZERO,
     $            WORK(IX), 1 )

C     Step 2: x=A'*x. p-1 times
      DO I = 1, P-1
         CALL DGEMV ( 'T', N, N, ONE, A, N, WORK(IX), 1, ZERO,
     $                WORK(IV), 1 )
C        x <--> v
         AUX = IX
         IX = IV
         IV = AUX
      END DO

      K = 2
      CONT = .TRUE.
      DO WHILE ( CONT )
C        [*,j] = max(abs(x)) 
         J = 1
         DO I = 2, N
            IF ( DABS( WORK(IX+I-1) ).GT.DABS( WORK(IX+J-1) ) ) J = I
         END DO

C        v = A^p*e_j
C        Step 1: v = A*e_j --> v=A(:,J)
         DO I = 1, N
            WORK(IV+I-1) = A(I, J)
         END DO

C        Step 2: v=A*v, p-1 times
         DO I = 1, P-1
            CALL DGEMV ( 'N', N, N, ONE, A, N, WORK(IV), 1, ZERO,
     $                   WORK(IX), 1 )
C           v <--> x
            AUX = IX
            IX = IV
            IV = AUX
         END DO

         OLD_NORM = NORM1P
         NORM1P = DLANGE( '1', N, 1, WORK(IV), 1, DUMMY )

C        Check if sign(v)=psi
         AUX = 0
         I = 1
         DO WHILE ( I.LE.N .AND. AUX.EQ.0 )
            DAUX1 = WORK(IPSI+I-1)
            DAUX2 = WORK(IV+I-1)
            IF ( (DAUX1.EQ.ZERO .AND. DAUX2.NE.ZERO) .OR.
     $           (DAUX1.EQ.ONE  .AND. DAUX2.LE.ZERO) .OR.
     $           (DAUX1.EQ.-ONE .AND. DAUX2.GE.ZERO) ) THEN
               AUX = 1
            END IF
            I = I+1
         END DO 

         IF ( AUX.EQ.0 .OR. NORM1P.LE.OLD_NORM ) THEN
            CONT = .FALSE.
         ELSE
C           Update psi = sign(v)      
            DO I = 1, N
               DAUX1 = WORK(IV+I-1)
               IF ( DAUX1.EQ.ZERO ) THEN
                  WORK(IPSI+I-1) = ZERO
               ELSEIF ( DAUX1.GT.ZERO ) THEN
                  WORK(IPSI+I-1) = ONE
               ELSE
                  WORK(IPSI+I-1) = -ONE
               END IF
            END DO

C           x = (A')^p*psi
C           Step 1:  x = A'*psi
            CALL DGEMV( 'T', N, N, ONE, A, N, WORK(IPSI), 1, ZERO,
     $                  WORK(IX), 1 )

C           Step 2: x = A'*x, p-1 times
            DO I = 1, P-1
               CALL DGEMV ( 'T', N, N, ONE, A, N, WORK(IX), 1, ZERO,
     $                      WORK(IV), 1 )
C              x <--> v
               AUX = IX
               IX = IV
               IV = AUX
            END DO

            K = K+1
            DAUX1 = DLANGE( 'M', N, 1, WORK(IX), 1, DUMMY )
            IF ( DAUX1.EQ.WORK(IX+J-1) .OR. K.GT.5 ) CONT = .FALSE.
         END IF
      END DO

      AUX = 1
      DO I = 1, N
C         WORK(IX+I-1) = (ONE + (I-ONE)/(N-ONE)) * (-1)**(I+1)
         WORK(IX+I-1) = (ONE + (I-ONE)/(N-ONE)) * AUX
         AUX = -AUX
      END DO

C     x = A^p*x
C     Step 1: x = A*x
      CALL DGEMV( 'N', N, N, ONE, A, N, WORK(IX), 1, ZERO,
     $            WORK(IV), 1 )
C     x <--> v
      AUX = IX
      IX = IV
      IV = AUX

C     Step 2:  x=A*x, p-1 times
      DO I = 1, P-1
         CALL DGEMV ( 'N', N, N, ONE, A, N, WORK(IX), 1, ZERO,
     $                WORK(IV), 1 )
C        x <--> v
         AUX = IX
         IX = IV
         IV = AUX
      END DO

      DAUX1 = 2*DLANGE( '1', N, 1, WORK(IX), 1, DUMMY )/(3*N)
      IF ( DAUX1.GT.NORM1P ) NORM1P = DAUX1
*
*     End of NORM1P
*
      END

